import logging
import subprocess
from pathlib import Path
from datetime import datetime
from .tables import zabbix
#from shlex import quote as shquote


def quote(s):
    # good enough? shlex.quote was adding too many extras
    r = repr(s)
    if (
        len(r) - 2 != len(s)
        or any(ch in s for ch in " ()|")
    ):
        return r
    
    return s


class CurryCommand(object):
    def __init__(self, cmd, env, env_extra):
        merged_env = {**env, **env_extra}
        self.cmd, self.env, self.env_extra = (
            list(map(str, cmd)),
            merged_env,
            env_extra,
        )

    def exec(self, *args):
        command = self.cmd + list(args)
        return run(command, env=self.env)

    def reprexec(self, *args):
        rargs = map(quote, self.cmd + list(args))

        str_env = " ".join((
            f"{key}={quote(value)}"
            for key, value
            in self.env_extra.items()))

        # Good enough
        output = ""
        if str_env:
            output += str_env + " \\\n"

        output += " ".join((
            f"\\\n    {line}" if line.startswith("-") else line for line in rargs
        ))
        
        return output

    __repr__ = reprexec


def run(*args, **kwargs):
    """
    Wrapper for subprocess.run.
    
    force 'text' output and returns 'stdout' as a tuple of lines
    where the last line is omitted if empty (it generally is).

    Return None on error (actual error, not the process retvalue)
    """
    if "text" not in kwargs:
        kwargs["text"] = True

    stderr = subprocess.PIPE
    try:
        logger = logging.getLogger()
        if logger.isEnabledFor(logging.DEBUG):
            stderr = None
    except Exception:
        pass

    try:
        result = subprocess.run(
            *args, **kwargs,
            check=True,
            stdout=subprocess.PIPE,
            stderr=stderr)
        out = result.stdout

    except subprocess.CalledProcessError as e:
        #logging.debug(e.stderr.rstrip())
        logging.debug(f"return code {e.returncode}")
        return None
    except FileNotFoundError:
        logging.critical(f"Command not found \"{args[0]}\"")
        return None

    lines = tuple(map(str.strip, out.split("\n")))
    if lines:
        if lines[-1] == "":
            return lines[:-1]

    return lines


def check_binary(*names):
    """Checks whether 'names' are all valid commands in the current shell."""
    out = run(("command", "-v", *names))

    if out is None:
        return False

    return len(out) == len(names)


def try_find_sockets(search, port):
    """Try to locate available postgresql sockets."""
    if not check_binary("netstat"):
        return tuple()

    out = run(("netstat", "-lxn"))
    sockets = []
    for line in out:
        # not perfect but it works reasonably enough
        try:
            path = Path(line.split()[-1])
        except IndexError:
            pass
            folder = path.parent
            name = path.name
            if search not in folder or str(port) not in name:
                continue

            sockets.append(path)

    return tuple(sockets)


def rlookup(ip):
    output = run(["dig", "+noall", "+answer", "-x", ip])
    if not output:
        return None

    parts = list(map(str.strip, output.split()))
    if len(parts) == 0:
        return None
    dn = parts[-1]

    subdomains = dn.split(".")
    if len(subdomains) == 0:
        return None

    return subdomains[0]


def parse_zabbix_version(query_result):
    raw_version = query_result[0]
    major = int(raw_version[:-6])
    minor = int(raw_version[-6:-4])
    revision = int(raw_version[-4:])
    
    version = f"{major}.{minor}.{revision}"

    return version, (major, minor, revision)


def create_name(args, version):
    dt = datetime.now().strftime("%Y%m%d-%H%M")
    name = f"zabbix_{args.host}_{dt}_{version}"

    return name


def preprocess_tables_lists(args, table_list):
    logging.debug(f"Table list: {table_list!r}")
    logging.verbose(f"Tables found: {len(table_list)}")

    tables = set(table_list)
    config = tables.intersection(zabbix.config)
    monitoring = tables.intersection(zabbix.monitoring)
    unknown = tables.difference(config, monitoring)

    logging.verbose(f"Config tables: {len(config)}")
    logging.verbose(f"Monitoring tables: {len(monitoring)}")
    logging.verbose(f"Unknown tables: {len(unknown)}")

    nodata, ignore, fail = [], [], []
    if args.monitoring == "nodata":
        nodata += monitoring

    if args.unknown == "nodata":
        nodata += unknown
    elif args.unknown == "ignore":
        ignore += unknown 
    elif args.unknown == "fail":
        fail += unknown

    return sorted(ignore), sorted(nodata), sorted(fail)


def pretty_log_args(args):
    """Print arguments via 'logging.info' in a readable way"""

    keys = args._keys

    str_args = ["Arguments:"]
    for key in keys:
        value = getattr(args, key, None)
        #if value is None:
        #    continue
        if key == "passwd" and value is not None:
            value = "[omissis]"

        str_args.append(f"    {key:<24}: {value}")

    #for key, value in dict_args.items():
    #    if key in keys or key == "scope":
    #        continue
    #    str_args.append(f"    {key:<24}: {value}")

    logging.verbose("\n".join(str_args))


import logging
from sys import exit
from os import environ
import atexit
from copy import deepcopy
from pathlib import Path
from .utils import (
    CurryCommand, parse_zabbix_version, create_name,
    preprocess_tables_lists, try_find_sockets,
)


def backup_mysql(args):
    logging.verbose(f"DBMS: MySQL or MariaSql")

    # Phase 0: setup authentication
    _mysql_auth(args)


    # Phase 1: Fetch database version
    select_db_version = "SELECT optional FROM dbversion;"
    raw_version = _mysql_query(args, select_db_version, "zabbix version query").exec()
    if raw_version is None:
        logging.fatal("Could not retrieve db version (see logs using --debug)")
        exit(1)

    version, _ = parse_zabbix_version(raw_version)
    logging.info(f"Zabbix version: {version}")

    name = create_name(args, version)
    logging.info(f"Backup base name: {name}")

    # Phase 2: Perform the actual backup
    dump_args = []
    
    # select and filter tables: done here and passed to _mysql_dump for simplicity
    table_list_query = (
        f"SELECT table_name FROM information_schema.tables "
        f"WHERE table_schema='{args.dbname}';")

    table_cmd = _mysql_query(args, table_list_query, "zabbix tables list query")
    table_list = sorted(table_cmd.exec())
    ignore, nodata, fail = preprocess_tables_lists(args, table_list)

    schema_path = Path(args.scope["tmp_dir"]) / f"schemas.sql"
    data_path = Path(args.scope["tmp_dir"]) / f"data.sql"

    dump_schema_args = [
        "--opt", "--single-transaction", "--skip-lock-tables",
        "--no-data", "--routines",
        "--result-file", schema_path, args.dbname,
        ]
    dump_data_args = [
        "--opt", "--single-transaction", "--skip-lock-tables",
        "--no-create-info", "--skip-extended-insert", "--skip-triggers",
        "--result-file", data_path, args.dbname,
    ]

    if fail:
        logging.error(f"Unknwon tables: aborting ({fail!r})")
        exit(1)

    if nodata:
        for table in nodata:
            dump_data_args += ["--ignore-table", f"{args.dbname}.{table}"]

    if ignore:
        for table in ignore:
            dump_schema_args += ["--ignore-table", f"{args.dbname}.{table}"]
            dump_data_args += ["--ignore-table", f"{args.dbname}.{table}"]

    dump_schema = _mysql_dump(args, dump_schema_args, f"Dump schema command", logging.info)
    dump_data = _mysql_dump(args, dump_data_args, f"Dump data command", logging.info)

    if not args.dry_run:
        dump_schema.exec()
        dump_data.exec()

    args.scope["name"], args.scope["version"] = name, version 


def _mysql_auth(args):
    args.scope["env"] = deepcopy(environ)
    args.scope["env_extra"] = {}

    if args.loginfile is not None:
        pass
    # Create temporary mylogin.cnf file
    elif args.passwd is not None:
        logincnf = Path(f"./mylogin.cnf")
        with logincnf.open("w") as fh:
            fh.writelines([
                f"[client]\r\n",
                f"password={args.passwd}\r\n",
            ])
        logincnf.chmod(0o600)
        if not args.keeploginfile:
            atexit.register(lambda: logincnf.unlink())
        args.loginfile = "./mylogin.cnf"


def _mysql_query(args, query, description="query", log_func=logging.debug):
    # mysql command will be used to inspect the database
    extra = []
    if args.read_mysql_config:
        extra += ["--defaults-file", args.mysql_config, ]

    if args.loginfile:
        extra += [f"--defaults-extra-file={args.loginfile}", ]

    if args.sock is None:
        extra += ["--host", args.host]
    else:
        extra += ["--socket", args.sock]

    query = CurryCommand(
        [
            "mysql",
            "--login-path=client",
        ] + extra + [
            "--user", args.user,
            "--port", args.port,
            "--database", args.dbname,
            "--skip-column-names",
            "--execute"
        ] + [query]
        , args.scope["env"], args.scope["env_extra"]
    )

    log_func(f"{description}: \n{query.reprexec()}")

    return query


def _mysql_dump(args, params, description="dump cmd", log_func=logging.debug):
    extra = []
    if args.read_mysql_config:
        extra += ["--defaults-file", args.mysql_config, ]

    if args.loginfile:
        extra += [f"--defaults-extra-file={args.loginfile}", ]

    if args.sock is None:
        extra += ["--host", args.host]
    else:
        extra += ["--socket", args.sock]

    extra_args = []
    if args.columns:
        extra_args += ["--complete-insert", "--quote-names", ]
    
    if args.verbosity in ("very", "debug"):
        extra_args += ["--verbose"]

    cmd = CurryCommand(
        [
            "mysqldump",
        ] + extra + [
            "--user", args.user,
            "--port", args.port,
        ] + extra_args + params,
        args.scope["env"], args.scope["env_extra"]
    )

    log_func(f"{description}: \n{cmd.reprexec()}")

    return cmd



from pathlib import Path
from .utils import run
from .utils import CurryCommand
from os import environ
from copy import deepcopy
from shutil import rmtree
import logging


def parse_save_files(files):
    """
    Read a list of files and directories.
    
    One item per line.
    Spaces are trimmed and blank lines or lines starting with a # are ignored.
    """
    with open(files, "r") as fh:
        for line in fh:
            line = line.strip()
            if line.startswith("#"):
                continue
            yield Path(line)


def save_files(args):
    """
    Copy files and directories as per user arguments.
    
    Default file list: (module) zabbixbackup/assets/files.
    """
    if args.save_files is True:
        files = args.files
        if files == "-":
            files = Path(__file__).parent / "assets" / "files"

        copy_files(files, args.scope["tmp_dir"] / "root")


def copy_files(save_files, base_dir):
    """
    Copy a list of files or directories in base_dir.

    Directory structure is replicated.
    """
    base_dir = base_dir.absolute()
    items = parse_save_files(save_files)

    for item in items:
        if not item.exists():
            logging.info(f"Filepath not found {item}, ignoring...")
            continue

        dest = base_dir / item.absolute().relative_to("/")
        if not dest.parent.exists():
            dest.parent.mkdir(parents=True)
            # TODO: copy permission on entire directory tree?

        if run(["cp", "-a", str(item), str(dest)]) is None:
            logging.warning(f"Cannot copy {item}, ignoring...")
        else:
            logging.info(f"Copying {item}")


def archive(args):
    """
    Create the actual archive file.

    Based on user arguments it will be compressed accordingly.
    """
    scope = args.scope
    tmp_dir: Path = args.scope["tmp_dir"]
    

    archive_dir = args.outdir / args.scope["name"]
    tmp_dir.rename(archive_dir)

    if scope["archive_command"] != None:
        archive_name = (
            archive_dir.parent /
            (archive_dir.name + scope["archive_extension"])
        )

        env = deepcopy(environ)
        compress = CurryCommand(
            scope["archive_command"] + [archive_name, archive_dir],
            env, scope["archive_env"]
            )
        logging.debug(f"Compress command: \n{compress.reprexec()}")

        result = compress.exec()
        if result is not None:
            rmtree(archive_dir)
    else:
        # Leave as plain directory
        pass



if __name__ == "__main__":
    import os
    from sys import argv
    import logging
    import tempfile
    from pathlib import Path
    
    from .parser import parse
    from .utils import pretty_log_args
    from .backup_postgre import backup_postgresql
    from .backup_mysql import backup_mysql
    from .archiver import save_files, archive
    from .rotation import rotate

    # Parse and preprocess cli arguments
    args = parse(argv[1:])
    scope = args.scope

    # Create temporary file for the log in the current directory
    raw_fh, tmp_log = tempfile.mkstemp(prefix="zabbixbackup_", suffix=".log", text=True, dir=".")
    tmp_log = Path(tmp_log)
    log_fh = os.fdopen(raw_fh, "w")
    logger = logging.getLogger()
    logger_handler = logging.StreamHandler(log_fh)
    logger.addHandler(logger_handler)

    # Create temporary output directory for the backup
    tmp_dir = tempfile.mkdtemp(prefix="zabbix_", dir=".")
    tmp_dir = Path(tmp_dir)
    scope["tmp_dir"] = tmp_dir

    logging.debug(f"Temporary log file: {tmp_log}")

    # Pretty print arguments as being parsed and processed
    pretty_log_args(args)

    if scope["dbms"] == "psql":
        backup_postgresql(args)
    elif scope["dbms"] == "mysql":
        backup_mysql(args)

    save_files(args)

    # Detach  file logger and save log into place 
    logger.removeHandler(logger_handler)
    log_fh.close()
    tmp_log.rename(tmp_dir / "dump.log")

    # Archive, compress and move the backup to the final destination
    archive(args)

    # Rotate backups
    rotate(args)



import logging
from sys import exit
from os import environ
import atexit
from copy import deepcopy
from pathlib import Path
from .utils import (
    CurryCommand, parse_zabbix_version, create_name,
    preprocess_tables_lists, try_find_sockets,
)


def backup_postgresql(args):
    logging.info(f"DBMS: Postgresql")

    # Phase 0: setup authentication
    _psql_auth(args)

    # Informational data about an eventual connection via socket
    if args.host == "" or args.host == "localhost" or args.host.startswith("/"):
        sockets = try_find_sockets("postgres", args.port)
        logging.info(f"sockets (actual choice performed directly by postgresql): ")
        logging.info(f"    {sockets!r}")


    # Phase 1: Fetch database version and assign a name
    select_db_version = "SELECT optional FROM dbversion;"
    raw_version = _psql_query(args, select_db_version, "zabbix version query").exec()
    if raw_version is None:
        logging.fatal("Could not retrieve db version (see logs using --debug)")
        exit(1)

    version, _ = parse_zabbix_version(raw_version)
    logging.info(f"Zabbix version: {version}")

    name = create_name(args, version)
    logging.info(f"Backup base name: {name}")


    # Phase 2: Perform the actual backup
    dump_args = []
    
    # select and filter tables: done here and passed to _pg_dump for simplicity
    table_list_query = (
        f"SELECT table_name FROM information_schema.tables "
        f"WHERE table_schema='{args.schema}' AND "
        f"table_catalog='{args.dbname}' AND "
        f"table_type='BASE TABLE';")

    table_cmd = _psql_query(args, table_list_query, "zabbix tables list query")
    table_list = sorted(table_cmd.exec())
    ignore, nodata, fail = preprocess_tables_lists(args, table_list)

    if fail:
        logging.error(f"Unknwon tables: aborting ({fail!r})")
        exit(1)

    if nodata:
        for i in range(0, len(nodata), 4):
            nodata_pattern = f"({'|'.join(nodata[i:i+4])})"
            dump_args += ["--exclude-table-data", nodata_pattern]

    if ignore:
        for i in range(0, len(ignore), 4):
            ignore_pattern = f"({'|'.join(ignore)})"
            dump_args += ["--exclude-table", ignore_pattern]

    # all other flags and arguments are set up by _pg_dump
    dump = _pg_dump(args, dump_args, "pgdump command", logging.info)

    if not args.dry_run:
        dump.exec()

    args.scope["name"], args.scope["version"] = name, version


def _psql_auth(args):
    args.scope["env"] = deepcopy(environ)
    args.scope["env_extra"] = {}

    if args.loginfile is not None:
        args.scope["env_extra"] = {"PGPASSFILE": str(args.loginfile)}
    elif args.passwd is not None:
        # Create temporary pgpass file
        pgpassfile = Path(f"./.pgpass")
        with pgpassfile.open("w") as fh:
            # TODO: socket?
            fh.write(f"{args.host}:{args.port}:{args.dbname}:{args.user}:{args.passwd}")
        pgpassfile.chmod(0o600)
        if not args.keeploginfile:
            atexit.register(lambda: pgpassfile.unlink())
        args.scope["env_extra"] = {"PGPASSFILE": str(pgpassfile)}


def _psql_query(args, query, description="query", log_func=logging.debug):
    # psql command will be used to inspect the database
    query = CurryCommand(
        [
            "psql",
            "--host", args.host,
            "--username", args.user,
            "--port", args.port,
            "--dbname", args.dbname,
            "--no-password", "--no-align", "--tuples-only", "--no-psqlrc",
            "--command",
        ] + [query]
        , args.scope["env"], args.scope["env_extra"]
    )

    log_func(f"{description}: \n{query.reprexec()}")

    return query


def _pg_dump(args, params, description="dump cmd", log_func=logging.debug):
    extra_args = []
    if args.columns:
        # Todo: figure out if --inserts is redundant
        extra_args += ["--inserts", "--column-inserts", "--quote-all-identifiers", ]
    
    if args.verbosity in ("very", "debug"):
        extra_args += ["--verbose"]

    if args.compression is not None:
        extra_args += ["--compress", args.compression]

    # choose the extension depending on output format
    # TODO: move out of pg_dump for simmetry
    extensions = {"plain": ".sql", "custom": ".pgdump", "directory": "", "tar": ".tar"}
    sqlpath = Path(args.scope["tmp_dir"]) / f"zabbix_cfg{extensions[args.format]}"

    extra_args += ["--file", str(sqlpath)]
    extra_args += ["--format", args.format]

    cmd = CurryCommand(
        [
            "pg_dump",
            "--host", args.host,
            "--username", args.user,
            "--port", args.port,
            "--dbname", args.dbname,
            "--schema", args.schema,
            "--no-password",
        ] + extra_args + params,
        args.scope["env"], args.scope["env_extra"]
    )

    log_func(f"{description}: \n{cmd.reprexec()}")

    return cmd
